from os.path import join, basename
from glob import glob
from random import random, choice, randrange, randint, shuffle
from math import degrees

from pygame import Surface
from pygame.draw import circle, polygon
from pygame.gfxdraw import aacircle
from pygame.image import load
from pygame.transform import rotate
from pygame.event import clear
from pygame.font import Font
from pygame.mixer import Sound, music
from pygame.locals import *

from lib.pgfw.pgfw.Game import Game
from lib.pgfw.pgfw.GameChild import GameChild
from lib.pgfw.pgfw.Sprite import Sprite
from lib.pgfw.pgfw.Vector import Vector
from lib.pgfw.pgfw.Animation import Animation
from lib.pgfw.pgfw.extension import get_distance, get_angle, get_delta, get_endpoint

class PlanetLicker(Game, Animation):

    SPAWN_MARGIN = 200
    LICK_A, LICK_B, LICK_C = "lick-a", "lick-b", "lick-c"

    def __init__(self):
        Game.__init__(self)
        Animation.__init__(self, self)
        self.sound_effects = SoundEffects(self)
        self.display_surface = self.get_display_surface()
        self.monster = Monster(self)
        self.planets = Planets(self)
        music.load(self.get_resource("Lick-Through.ogg"))
        music.play(-1)
        music.set_volume(.9)
        instructions = self.instructions = Sprite(self)
        instructions.add_frame(load(self.get_resource("movement.png")).convert_alpha())
        instructions.location.center = self.display_surface.get_rect().center
        instructions.location.top += 60
	instructions.hide()
        self.subscribe(self.respond)
        self.register(self.reset, self.unsuppress_commands)
        self.reset()

    def set_background(self):
        field = []
        w, h = [Planets.WORLD_SIZE] * 2
        for x in xrange(w):
            field.append([])
            for y in xrange(h):
                if random() < .0005:
                    field[-1].append((randrange(0, 360), choice((10, 20, 30, 40, 50))))
                else:
                    field[-1].append(None)
        background = self.background = Sprite(self, 800)
        color = Color(0, 0, 0)
        for _ in xrange(8):
            frame = Surface((w, h))
            frame.fill((0, 0, 0))
            for x, row in enumerate(field):
                for y, cell in enumerate(row):
                    if cell:
                        rect = Rect((0, 0), [randint(1, 4)] * 2)
                        rect.center = x, y
                        if random() < .5:
                            color.hsla = (cell[0] + randint(-cell[1], cell[1])) % 360, \
                                         100, 50, 100
                        else:
                            color.hsla = 0, 100, 100, 100
                        for ox, oy in (0, 0), (-w, 0), (0, -h), (w, 0), (0, h):
                            frame.fill(color, rect.move(ox, oy))
                            frame.fill(color, rect.move(-1 + ox, -1 + oy))
                            frame.fill(color, rect.move(1 + ox, 1 + oy))
                            frame.fill(color, rect.move(-1 + ox, 1 + oy))
                            frame.fill(color, rect.move(1 + ox, -1 + oy))
                            frame.fill(color, rect.move(-2 + ox, 0 + oy))
                            frame.fill(color, rect.move(2 + ox, 2 + oy))
                            frame.fill(color, rect.move(2 + ox, 0 + oy))
                            frame.fill(color, rect.move(-2 + ox, -2 + oy))
            background.add_frame(frame)
        background.add_location(offset=(w, 0))
        background.add_location(offset=(0, h))
        background.add_location(offset=(w, h))

    def respond(self, event):
        if not self.commands_suppressed and not self.game_over:
            compare = self.get_game().delegate.compare
            if compare(event, "reset-game"):
                self.reset()
            elif self.compare_lick(event, self.LICK_A):
                self.add_vector(self.LICK_A)
                self.instructions.hide()
            elif self.compare_lick(event, self.LICK_B):
                self.add_vector(self.LICK_B)
                self.instructions.hide()
            elif self.compare_lick(event, self.LICK_C):
                self.add_vector(self.LICK_C)
                self.instructions.hide()

    def unsuppress_commands(self):
        self.commands_suppressed = False

    def compare_lick(self, event, lick):
        compare = self.get_game().delegate.compare
        if self.check_command_line("-makey"):
            return (lick == self.LICK_A and compare(event, "left")) or \
                (lick == self.LICK_B and compare(event, "up")) or \
                (lick == self.LICK_C and compare(event, "right"))
        else:
            return compare(event, lick)

    def add_vector(self, lick):
        closest = None
        min_distance = None
        for planet in self.planets.planets:
            if planet.lick == lick:
                for location in planet.locations:
                    distance = get_distance(location.center, self.monster.location.center)
                    if min_distance is None or distance < min_distance:
                        closest = location
                        min_distance = distance
        if closest is None:
            self.get_game().sound_effects.play("stall")
        else:
            planets = self.monster.collide_with_planets()
            if planets and lick in [planet.lick for planet in planets]:
                for planet in planets:
                    if planet.lick == lick:
                        break
                self.monster.show_lick()
                planet.shrink()
            else:
                angle = get_angle(closest.center, self.monster.location.center)
                vector = Motion(self, degrees(angle))
                self.vectors.append(vector)
                self.get_game().sound_effects.play("propell-2")

    def reset(self):
        self.time_filter.close()
        self.game_over = False
        self.set_background()
        self.planets.reset()
        self.monster.reset()
        self.halt()
        self.obstacles = []
        self.vectors = []
        while True:
            conflict = False
            self.move(randint(0, Planets.WORLD_SIZE),
                      randint(0, Planets.WORLD_SIZE))
            for planet in self.planets.planets:
                for location in planet.locations:
                    if get_distance(location.center, self.monster.location.center) < \
                       self.SPAWN_MARGIN:
                        conflict = True
                        break
                if conflict:
                    break
            if not conflict:
                break
	if not self.check_command_line("-makey"):
            self.instructions.unhide()
        self.commands_suppressed = True
        clear()
        self.monster.get_current_frameset().reset()
        self.time_filter.open()
        self.reset_timer(self.unsuppress_commands)
        self.play(self.unsuppress_commands, delay=2000, play_once=True)

    def move(self, x, y):
        w, h = self.background.location.size
        self.background.move(-x, -y)
        if self.background.location.top > 0:
            self.background.move(dy=-h)
        if self.background.location.right < 0:
            self.background.move(w)
        if self.background.location.bottom < 0:
            self.background.move(dy=h)
        if self.background.location.left > 0:
            self.background.move(-w)
        for planet in self.planets.planets:
            planet.move(-x, -y)
            if planet.location.top > 0:
                planet.move(dy=-h)
            if planet.location.right < 0:
                planet.move(w)
            if planet.location.bottom < 0:
                planet.move(dy=h)
            if planet.location.left > 0:
                planet.move(-w)
        for obstacle in self.obstacles:
            obstacle.move(-x, -y)
            if obstacle.location.top > 0:
                obstacle.move(dy=-h)
            if obstacle.location.right < 0:
                obstacle.move(w)
            if obstacle.location.bottom < 0:
                obstacle.move(dy=h)
            if obstacle.location.left > 0:
                obstacle.move(-w)

    def add_obstacles(self, origin):
        for _ in xrange(6):
            obstacle = Sprite(self, 40)
            color = Color(0, 0, 0)
            for color in ((255, 255, 255), (255, 255, 0), (128, 255, 0),
                          (255, 128, 0), (255, 0, 0), (255, 255, 255), (255, 0, 0)):
                frame = Surface((19, 19), SRCALPHA)
                radius = frame.get_width() / 2
                circle(frame, color, [radius] * 2, radius)
                color = Color(*color)
                h, s, l, a = color.hsla
                color.hsla = (h + 30) % 60, s, l, a
                circle(frame, color, [radius] * 2, radius - 2)
                frame.fill((255, 0, 0), (0, radius - 2, radius * 2, 5))
                frame.fill((0, 0, 0, 0), (0, radius - 1, radius * 2, 3))
                obstacle.add_frame(frame)
            while True:
                obstacle.location.center = get_endpoint(origin, randrange(0, 360),
                                                        randint(150, 280))
                conflict = False
                for planet in self.planets.planets:
                    if planet.location.colliderect(obstacle.location):
                        conflict = True
                        break
                for other in self.obstacles:
                    if other.location.colliderect(obstacle.location):
                        conflict = True
                        break
                if not conflict:
                    break
            obstacle.add_location(offset=(Planets.WORLD_SIZE, 0))
            obstacle.add_location(offset=(0, Planets.WORLD_SIZE))
            obstacle.add_location(offset=(Planets.WORLD_SIZE, Planets.WORLD_SIZE))
            self.obstacles.append(obstacle)

    def end_game(self, success=False):
        self.game_over = True
        font = Font(self.get_resource("PetMe.ttf"), 64)
        color = (255, 0, 0)
        if success:
            frame = font.render("YOU WIN!", True, (0, 255, 0))
            self.get_game().sound_effects.play("win")
        else:
            frame = font.render("GAME OVER", True, (255, 0, 0))
        sprite = self.game_over_sprite = Sprite(self)
        sprite.add_frame(frame)
        sprite.location.center = self.display_surface.get_rect().center
        self.play(self.reset, delay=5000, play_once=True)

    def update(self):
        Animation.update(self)
        for vector in self.vectors:
            self.move(*vector.delta)
            vector.update()
            if vector.delta == (0, 0):
                self.vectors.remove(vector)
        if not self.vectors and self.monster.get_current_frameset().name != "lick":
            self.monster.set_frameset("idle")
        elif self.monster.get_current_frameset().name != "lick":
            self.monster.set_frameset("moving")
        self.background.update()
        self.planets.update()
        damaged = False
        for obstacle in self.obstacles:
            for location in obstacle.locations:
                if location.colliderect(self.monster.location) and \
                   get_distance(location.center, self.monster.location.center) < \
                   location.w / 2 + self.monster.location.w / 2:
                    if not damaged:
                        self.monster.damage()
                        damaged = True
                    self.obstacles.remove(obstacle)
            obstacle.update()
        self.show_hints()
        self.monster.update()
        if self.game_over:
            self.game_over_sprite.update()
        self.instructions.update()

    def show_hints(self):
        for lick in Planets.LICKS:
            closest = None
            min_distance = None
            for planet in self.planets.planets:
                for location in planet.locations:
                    if planet.lick == lick:
                        distance = get_distance(self.monster.location.center,
                                                location.center)
                        if not closest or distance < min_distance:
                            color = planet.get_color()
                            closest = location
                            min_distance = distance
            rect = self.get_display_surface().get_rect()
            if closest and not closest.colliderect(rect.inflate(-20, -20)):
                surface = Surface((20, 30))
                surface.fill((255, 255, 255))
                transparent = (255, 0, 255)
                surface.set_colorkey(transparent)
                polygon(surface, transparent, ((0, 10), (0, 0), (10, 0)))
                polygon(surface, transparent, ((10, 0), (20, 0), (20, 10)))
                circle(surface, color, (10, 20), 8)
                if closest.centery < 0 or closest.centery > rect.h:
                    x = closest.centerx
                    if x < 0:
                        x = 0
                    elif x > rect.w - surface.get_width():
                        x = rect.w - surface.get_width()
                    if closest.centery < 0:
                        y = 0
                    else:
                        y = rect.h - surface.get_height()
                        surface = rotate(surface, 180)
                else:
                    y = closest.centery
                    if closest.centery < 0:
                        y = 0
                    elif closest.centery > rect.h - surface.get_height():
                        y = rect.h - surface.get_height()
                    if closest.centerx < 0:
                        x = 0
                        surface = rotate(surface, 90)
                    else:
                        x = rect.w - surface.get_height()
                        surface = rotate(surface, 270)
                self.get_display_surface().blit(surface, (x, y))


class SoundEffects(GameChild):

    def __init__(self, parent):
        GameChild.__init__(self, parent)
        effects = self.effects = []
        for path in glob(join(self.get_resource("sfx"), "*.wav")):
            effects.append(SoundEffect(self, path))

    def play(self, name, volume=None):
        for effect in self.effects:
            if effect.name == name:
                effect.play(volume=volume)


class SoundEffect(GameChild, Sound):

    DEFAULT_VOLUME = 1.0

    def __init__(self, parent, path, volume=DEFAULT_VOLUME):
        GameChild.__init__(self, parent)
        Sound.__init__(self, path)
        self.name = basename(path).split(".")[0]
        self.display_surface = self.get_display_surface()
        self.set_volume(volume)

    def play(self, loops=0, maxtime=0, fade_ms=0, position=None, x=None, volume=None):
        channel = Sound.play(self, loops, maxtime, fade_ms)
        if x is not None:
            position = float(x) / self.display_surface.get_width()
	if position is not None and channel is not None:
            channel.set_volume(*self.get_panning(position))
        return channel

    def get_panning(self, position):
        return 1 - max(0, ((position - .5) * 2)), \
               1 + min(0, ((position - .5) * 2))


class Motion(GameChild):

    def __init__(self, parent, angle):
        GameChild.__init__(self, parent)
        self.angle = angle
        self.nodeset = self.get_game().interpolator.get_nodeset("speed")
        self.time_left = self.nodeset.get_length()
        self.update()

    def update(self):
        if self.time_left > 0:
            self.time_left -= self.get_game().time_filter.get_last_frame_duration()
            dx, dy = get_delta(self.angle, self.nodeset.get_y(self.time_left))
            self.delta = -dx, dy
        else:
            self.delta = 0, 0


class Monster(Sprite):

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.display_surface = self.get_display_surface()
        self.add_frames()
        self.set_life_meter_sprites()
        self.location.center = self.display_surface.get_rect().center
        self.register(self.end_lick)

    def add_frames(self):
        self.load_from_path(self.get_resource("biomass/"), True)
        self.add_frameset([0, 3, 4, 0, 5], [5000, 800, 800, 5000, 200], "idle", True)
        self.add_frameset([0, 1, 2, 1], 100, "lick")
        self.add_frameset(0, name="moving")
        # self.add_frameset([3, 4], [400, 800], "shifty")
        # self.add_frameset(5, name="cute")

    def set_life_meter_sprites(self):
        sprites = self.life_meter_sprites = []
        for path in sorted(glob(join(self.get_resource("life"), "*.png"))):
            sprite = Sprite(self)
            image = load(path).convert()
            image.set_colorkey((0, 0, 0))
            image.set_alpha(5)
            sprite.add_frame(image)
            sprite.location.bottomleft = self.get_display_surface().get_rect().move(10, -10).bottomleft
            sprite.set_alpha(255)
            sprites.append(sprite)

    def reset(self):
        self.life = 3
        self.halt(self.end_lick)
        self.set_frameset("idle")

    def collide_with_planets(self):
        planets = []
        for planet in self.get_game().planets.planets:
            for location in planet.locations:
                if location.colliderect(self.location) and \
                   get_distance(self.location.center, location.center) < \
                   location.w / 2 + self.location.w / 2:
                    planets.append(planet)
        return planets

    def damage(self):
        self.life -= 1
        if self.life == 0:
            self.get_game().end_game()
            self.get_game().sound_effects.play("player_death")
        else:
            self.get_game().sound_effects.play("take_damage-1")

    def show_lick(self):
        if not self.get_current_frameset().name == "lick":
            self.set_frameset("lick")
            self.play(self.end_lick, delay=800, play_once=True)
            self.get_game().sound_effects.play("licking")

    def end_lick(self):
        self.set_frameset("idle")

    def update(self):
        Sprite.update(self)
        self.life_meter_sprites[self.life].update()


class Planets(GameChild):

    WORLD_SIZE = 1500
    RADII = 64
    LICKS = PlanetLicker.LICK_A, PlanetLicker.LICK_B, PlanetLicker.LICK_C
    CITY, VOLCANO, UNDERWATER, BEACH, ARCADE, GHOST = xrange(6)
    MIN_MARGIN = 400

    def __init__(self, parent):
        GameChild.__init__(self, parent)

    def reset(self):
        self.populate()

    def populate(self):
        kinds = range(6)
        kinds_copy = list(kinds)
        for _ in xrange(3):
            kinds.append(kinds_copy.pop())
        radii = list([self.RADII] * 9)
        shuffle(radii)
        licks = list(self.LICKS + self.LICKS + self.LICKS)
        planets = self.planets = []
        for kind, radius, lick in zip(kinds, radii, licks):
            while True:
                conflict = False
                planet = Planet(self, radius, lick, kind)
                planet.location.center = randint(0, Planets.WORLD_SIZE), \
                                         randint(0, Planets.WORLD_SIZE)
                planet.add_location(offset=(Planets.WORLD_SIZE, 0))
                planet.add_location(offset=(0, Planets.WORLD_SIZE))
                planet.add_location(offset=(Planets.WORLD_SIZE, Planets.WORLD_SIZE))
                for other in self.planets:
                    for location in other.locations:
                        for my_location in planet.locations:
                            if get_distance(location.center, my_location.center) < self.MIN_MARGIN:
                                conflict = True
                                break
                        if conflict:
                            break
                    if conflict:
                        break
                if not conflict:
                    break
            planets.append(planet)

    def update(self):
        for planet in self.planets:
            planet.update()


class Planet(Sprite):

    COLORS = (251, 186, 68), (60, 133, 255), (60, 219, 90)
    SHRINK_STEP = 16
    CANVAS_SIZE = 160

    def __init__(self, parent, radius, lick, kind):
        Sprite.__init__(self, parent)
        self.max_radius = radius
        self.radius = radius
        self.lick = lick
        self.kind = kind
        self.set_ring()
        self.add_frames()

    def set_ring(self):
        ring = self.ring = None
        if self.kind == Planets.CITY:
            self.ring = load(self.get_resource("planets/city_128.png")).convert_alpha()
        elif self.kind == Planets.UNDERWATER:
            self.ring = load(self.get_resource("planets/underwater_128.png")).convert_alpha()
        elif self.kind == Planets.GHOST:
            self.ring = load(self.get_resource("planets/ghost_128.png")).convert_alpha()
        elif self.kind == Planets.BEACH:
            self.ring = load(self.get_resource("planets/beach_128.png")).convert_alpha()
        elif self.kind == Planets.VOLCANO:
            self.ring = load(self.get_resource("planets/volcano_128.png")).convert_alpha()
        elif self.kind == Planets.ARCADE:
            self.ring = load(self.get_resource("planets/arcade_128.png")).convert_alpha()

    def add_frames(self):
        self.clear_frames()
        frame = Surface([self.CANVAS_SIZE] * 2, SRCALPHA)
        circle(frame, self.get_color(), [self.CANVAS_SIZE / 2] * 2, self.radius)
        if self.ring:
            frame.blit(self.ring, (0, 0))
        self.add_frame(frame)

    def get_color(self):
        return self.COLORS[Planets.LICKS.index(self.lick)]

    def shrink(self):
        self.radius -= self.SHRINK_STEP
        if self.radius == 0:
            self.get_game().add_obstacles(self.location.center)
            self.get_game().sound_effects.play("planet_eaten-1")
            self.parent.planets.remove(self)
            if len(self.parent.planets) == 3:
                self.get_game().end_game(True)

    def update(self):
        self.add_frames()
        Sprite.update(self)
